import matplotlib.pyplot as plt
import numpy as np
import torch
from drw_tools import plot_cf
import os


def r2w(z_r, x):
    z_w = torch.zeros_like(x)
    for t in range(x.shape[0]):
        R = torch.vstack((torch.hstack((torch.cos(x[t, 2]), -torch.sin(x[t, 2]))),torch.hstack((torch.sin(x[t, 2]), torch.cos(x[t, 2])))))
        if z_w.shape[0] == 1:
            z_w[t, 0:2] = R @ z_r[0, 0:2] + x[t, 0:2]
            z_w[t, 2] = z_r[0, 2] + x[t, 2]
        else:
            z_w[t, 0:2] = R @ z_r[t, 0:2] + x[t, 0:2]
            z_w[t, 2] = z_r[t, 2] + x[t, 2]
    return z_w


def w2r(z_w, x):
    z_r = torch.zeros_like(x)
    for t in range(x.shape[0]):
        R = torch.vstack((torch.hstack((torch.cos(x[t, 2]), -torch.sin(x[t, 2]))), torch.hstack((torch.sin(x[t, 2]), torch.cos(x[t, 2])))))
        if z_w.shape[0] == 1:
            z_r[t, 0:2] = R.t() @ (z_w[0, 0:2] - x[t, 0:2])
            z_r[t, 2] = z_w[0, 2] - x[t, 2]
        else:
            z_r[t, 0:2] = R.t() @ (z_w[t, 0:2] - x[t, 0:2])
            z_r[t, 2] = z_w[t, 2] - x[t, 2]
    return z_r


def simulate(mr_gt, ma_gt, traj_length=6, max_dist=3, zmr_noise=0.1, zma_noise=0.1, zodom_noise=0.3):
    def r2w_np(z_r, x):
        z_w = np.zeros(3)
        R = np.vstack((np.hstack((np.cos(x[2]), -np.sin(x[2]))), np.hstack((np.sin(x[2]), np.cos(x[2])))))
        z_w[0:2] = R @ z_r[0:2] + x[0:2]
        z_w[2] = z_r[2] + x[2]
        return z_w

    def w2r_np(z_w, x):
        z_r = np.zeros(3)
        R = np.vstack((np.hstack((np.cos(x[2]), -np.sin(x[2]))), np.hstack((np.sin(x[2]), np.cos(x[2])))))
        z_r[0:2] = R.transpose() @ (z_w[0:2] - x[0:2])
        z_r[2] = z_w[2] - x[2]
        return z_r

    # control (ground truth)
    u = 12.5 / traj_length * np.ones([traj_length, 2])
    u[:, 1] = np.pi / 0.5 / (1*traj_length)
    dk = 1.0

    # initial state (ground truth)
    x, y, yaw = 1.0, 0.0, 0.0

    # GENERATE MEASUREMENTS
    zmr = []   # measurements of relative marker
    zma = []    # measurements of absolute marker
    zodom = []  # measurement of relative motion
    x_gt = []   # ground truth trajectory
    for k in range(u.shape[0]):
        v_lin = u[k, 0]
        v_yaw = u[k, 1]
        x_old = np.array([x, y, yaw])
        x, y, yaw = x + v_lin * np.cos(yaw) * dk, y + v_lin * np.sin(yaw) * dk, yaw + v_yaw * dk
        x_gt.append([x, y, yaw])

        # zodom_aux = np.zeros([1, 3])
        zodom_aux = w2r_np(np.array([x, y, yaw]), x_old) + zodom_noise * (np.random.rand(3) - 0.5)
        zodom.append(zodom_aux)

        zmr_aux = np.zeros([mr_gt.shape[0], 3])
        for i, (m_i) in enumerate(mr_gt):
            zmr_aux[i, :] = w2r_np(mr_gt[i, :], [x, y, yaw]) + zmr_noise * (np.random.rand(3) - 0.5)
        if np.linalg.norm(zmr_aux[0, 0:2]) > max_dist:
            zmr_aux = np.full(zmr_aux.shape, np.nan)
        zmr.append(zmr_aux)

        zma_aux = np.zeros([ma_gt.shape[0], 3])
        for i, (m_i) in enumerate(ma_gt):
            zma_aux[i, :] = w2r_np(ma_gt[i, :], [x, y, yaw]) + zma_noise * (np.random.rand(3) - 0.5)
        if np.linalg.norm(zma_aux[0, 0:2]) > max_dist:
            zma_aux = np.full(zma_aux.shape, np.nan)
        zma.append(zma_aux)
    return x_gt, u, zodom, zma, zmr


def visu(x, mr, ma, mr_gt, ma_gt, zmr, zma, t):
    zmr_w = r2w(zmr, x).detach().numpy()
    zma_w = r2w(zma, x).detach().numpy()
    x_np, mr_np, ma_np = x.detach().numpy(), mr.detach().numpy(), ma.detach().numpy()

    ax.clear()
    plt.grid(True)

    # marker measurements
    for k in range(t):
        ax.plot([x_np[k, 0], zma_w[k, 0]], [x_np[k, 1], zma_w[k, 1]], 'x-', color='red', markersize=10, mew=3)
        ax.plot([x_np[k, 0], zmr_w[k, 0]], [x_np[k, 1], zmr_w[k, 1]], 'x-', color='gray', markersize=10, mew=3)

    # estimated robot pose
    ax.plot(x_np[0:t, 0], x_np[0:t, 1], 'o', color='white', markersize=12, mew=3, zorder=20)
    ax.plot(x_np[0:t, 0], x_np[0:t, 1], 'o', color='magenta', markersize=10, mew=3, zorder=20)
    for k in range(t):
        plot_cf(x_np[k, :], length=0.4)

    # estimated marker pose
    ax.plot(mr_np[:, 0], mr_np[:, 1], 'go', markersize=12, mew=3, zorder=20)
    ax.plot(ma_np[:, 0], ma_np[:, 1], 'bo', markersize=10, mew=3, zorder=20)
    plot_cf(ma_np[0, :], length=0.8)
    plot_cf(mr_np[0, :], length=0.8)

    for j in range(t):
        ax.plot(x_gt[j][0], x_gt[j][1], 'b.', markersize=3, mew=2)
        ax.plot(x_np[j, 0], x_np[j, 1], 'r.', markersize=3, mew=2)

    # ground truth marker position
    ax.plot(mr_gt[0, 0], mr_gt[0, 1], 'gx', markersize=10, mew=3)
    ax.plot(ma_gt[0, 0], ma_gt[0, 1], 'bx', markersize=10, mew=3)
    plot_cf(ma_gt[0, :], length=0.8)
    plot_cf(mr_gt[0, :], length=0.8)

    ax.set_aspect("equal", "box")
    plt.xlim(-2, 5)
    plt.ylim(-2, 5)
    plt.pause(0.001)


def get_measurements():
    idx_zma = np.where(~np.isnan(zma_all).any(axis=1) & (np.arange(zma_all.shape[0]) < t) & (np.arange(zma_all.shape[0]) > t - max_history))[0]
    idx_zmr = np.where(~np.isnan(zmr_all).any(axis=1) & (np.arange(zmr_all.shape[0]) < t) & (np.arange(zmr_all.shape[0]) > t - max_history))[0]
    idx_zodom = np.where(~np.isnan(zodom_all).any(axis=1) & (np.arange(zodom_all.shape[0]) < t) & (np.arange(zodom_all.shape[0]) > t - max_history))[0]
    zma = zma_all[idx_zma, :]
    zmr = zmr_all[idx_zmr, :]
    zodom = zodom_all[idx_zodom, :]
    return zma, zmr, zodom, idx_zma, idx_zmr, idx_zodom


# TODO PLAY WITH PARAMETERS
#  - measurement noise (zmr_noise, zma_noise, zodom_noise)
#  - maximal visibility distance (max_dist)
#  - marker positions (ma_gt, mr_gt)

# Ground truth markers in wcf (x,y,yaw)
ma_gt = np.array([[-1.0, 4.0, 0.0]])
mr_gt = np.array([[0.0, -1.0, 0.0]])

# Collect measurements and ground truth trajectory
x_gt, u, zodom_all, zma_all, zmr_all = simulate(mr_gt, ma_gt, traj_length=20, max_dist=3, zmr_noise=0.3, zma_noise=0.1, zodom_noise=0.1)
zma_all, zmr_all, zodom_all = torch.tensor(np.array(zma_all)).squeeze(), torch.tensor(np.array(zmr_all)).squeeze(), torch.tensor(np.array(zodom_all)[1:, :]).squeeze()

# Initialize
x, mr, ma = torch.tensor(1.2+np.array(x_gt)+0.3*(np.random.rand(len(x_gt), 3)-0.5), requires_grad=True), torch.tensor(np.array(mr_gt)+2, requires_grad=True), torch.tensor(np.array(ma_gt), requires_grad=True)

max_history = 100
img = 0
fig, ax = plt.subplots()

opt = torch.optim.Adam([x, mr], lr=0.1)
for t in range(2, x.shape[0], 1):

    for i in range(50):
        zma, zmr, zodom, idx_zma, idx_zmr, idx_zodom = get_measurements()

        loss = ...  # <------
        # TODO FILL IN
        # Hint0: In each time only a subset of measurements is available.
        #        States corresponding to available measurements are provided in "idx_zma", "idx_zmr", "idx_zodom".
        #        For example measurements "zma" corresponds to states "x[idx_zma, :]" etc...
        # Hint1: Absolute marker position in all rcfs is "w2r(ma, x[idx_zma, :])";
        #        therefore you want to have it as close as possible to "zma".
        #        Similarly "w2r(x[idx_zodom + 1, :], x[idx_zodom, :])" is relative motion
        #        between all consecutive robot poses;
        #        therefore you want to have it as close as possible to "zodom", etc...
        # Hint2: In order to compute squared distance between vectors a, b
        #        the following notation can be used in pytorch:
        #        ((a-b)**2).sum()
        #
        loss.backward()
        if t - max_history + 1 >= 0:
            x.grad[t - max_history + 1, :] = 0
        opt.step()
        opt.zero_grad()

        print('loss=', loss.detach().numpy())

    # visu
    visu(x, mr, ma, mr_gt, ma_gt, zmr_all, zma_all, t)

